from unittest import TestCase
from tempfile import mkdtemp
from shutil import rmtree, copyfile
from os import mkdir, listdir
from os.path import join, realpath, dirname, isdir, abspath
from v import main, c_ante, c_desc, c_flow, c_origin, c_roots, c_wot

class ArgMock(object):

    query = "bar"

class VIntegrationTests(TestCase):

    def setUp(self):
        self.scratchdir = mkdtemp()
        self.loc = dirname(abspath(__file__))
        self.patchdir = join(self.scratchdir, "patches")
        self.sealsdir = join(self.scratchdir, "seals")
        self.wotdir = join(self.scratchdir, "wot")
        self.args = ArgMock()
        self.args.vpatches = self.patchdir
        self.args.seals = self.sealsdir
        self.args.wot = self.wotdir
        for d in [self.patchdir, self.sealsdir, self.wotdir]:
            mkdir(d)

    def v_main(self, verbose="squelch", fingers=None, wild=False,
               cmd=c_wot, query="genesis.vpatch"):
        self.args.verbose = verbose
        self.args.fingers = fingers
        self.args.wild = wild
        self.args.f = cmd
        self.args.query = query
        return main(self.args)

    def copy_wotfile(self, wotfile):
        copyfile(join(self.loc, "wot", wotfile),
                 join(self.wotdir, wotfile))

    def copy_sigfile(self, sigfile):
        copyfile(join(self.loc, "sigs", sigfile),
                 join(self.sealsdir, sigfile))

    def copy_patchfile(self, patchname):
        copyfile(join(self.loc, "patches", patchname),
                 join(self.patchdir, patchname))

    def athenize(self):
        """spring forth from zeus!"""
        self.copy_patchfile("genesis.vpatch")
        self.copy_sigfile("genesis.vpatch.asciilifeform.sig")
        self.copy_wotfile("asciilifeform.asc")

    def orchestrate(self):
        for s in listdir(join(self.loc, "sigs")):
            self.copy_sigfile(s)
        for w in listdir(join(self.loc, "wot")):
            self.copy_wotfile(w)
        for p in listdir(join(self.loc, "patches")):
            self.copy_patchfile(p)

    def test_tooling_setup(self):
        for d in [self.patchdir, self.sealsdir, self.wotdir]:
            self.assertTrue(isdir(d))

    def test_roots_fails_with_no_patches(self):
        with self.assertRaises(SystemExit):
            self.v_main()

    def test_roots_fails_with_no_seals_or_wot(self):
        self.copy_patchfile("genesis.vpatch")
        with self.assertRaises(SystemExit):
            self.v_main()

    def test_roots_fails_with_a_seal_but_no_wot(self):
        self.copy_patchfile("genesis.vpatch")
        self.copy_sigfile("genesis.vpatch.asciilifeform.sig")
        with self.assertRaises(SystemExit):
            self.v_main()

    def test_roots_works_with_a_patch_a_seal_and_one_wot_member(self):
        self.athenize()
        raised = False
        try:
            n = self.v_main()
        except:
            raised = True
        self.assertFalse(raised)
        self.assertEquals(1, len(n))
        self.assertEquals(
            n,
            [{'fp': u'17215D118B7239507FAFED98B98228A001ABFFC7',
              'handle': 'asciilifeform',
              'id': 'Stanislav Datskovskiy <stas@loper-os.org>'}])

    def test_roots_works_with_a_patch_and_two_seals_and_members(self):
        self.copy_patchfile("genesis.vpatch")
        self.copy_sigfile("genesis.vpatch.asciilifeform.sig")
        self.copy_sigfile("genesis.vpatch.mircea_popescu.sig")
        self.copy_wotfile("asciilifeform.asc")
        self.copy_wotfile("mircea_popescu.asc")
        n = self.v_main()
        self.assertEquals(
            n,
            [{'fp': u'17215D118B7239507FAFED98B98228A001ABFFC7',
              'handle': 'asciilifeform',
              'id': 'Stanislav Datskovskiy <stas@loper-os.org>'},
             {'fp': u'6160E1CAC8A3C52966FD76998A736F0E2FB7B452',
              'handle': 'mircea_popescu',
              'id': 'Mircea Popescu (Acest articol are apriori avantajul aliteralitatii alaturi.) <office@polimedia.us>'}])

    def test_full_orchestras_roots(self):
        self.orchestrate()
        n = self.v_main()
        self.assertEquals(
            n,
            [{'fp': u'027A8D7C0FB8A16643720F40721705A8B71EADAF',
              'handle': 'mod6', 'id': 'mod6 (mod6) <modsix@gmail.com>'},
             {'fp': u'4F7907942CA8B89B01E25A762AFA1A9FD2D031DA',
              'handle': 'ben_vulpes',
              'id': 'Ben Vulpes <benkay@gmail.com>'},
             {'fp': u'6160E1CAC8A3C52966FD76998A736F0E2FB7B452',
              'handle': 'mircea_popescu',
              'id': 'Mircea Popescu (Acest articol are apriori avantajul aliteralitatii alaturi.) <office@polimedia.us>'},
             {'fp': u'FC66C0C5D98C42A1D4A98B6B42F9985AFAB953C4',
              'handle': 'trinque',
              'id': 'Michael Trinque <mike@trinque.org>'},
             {'fp': u'17215D118B7239507FAFED98B98228A001ABFFC7',
              'handle': 'asciilifeform',
              'id': 'Stanislav Datskovskiy <stas@loper-os.org>'},
             {'fp': u'F28E0095843B91CB22E7D65533588BE08B232B13',
              'handle': 'punkman', 'id': 'punkman'}])

    def test_roots_returns_signatories_to_a_patch(self):
        self.copy_patchfile("genesis.vpatch")
        self.copy_sigfile("genesis.vpatch.asciilifeform.sig")
        self.copy_wotfile("asciilifeform.asc")
        n = self.v_main()
        self.assertEquals(1, len(n))
        # TODO: test for set equality here

    def test_wot_fails_without_wot(self):
        with self.assertRaises(SystemExit):
            self.v_main()

    def test_lists_a_wot_member(self):
        sigfile = listdir(join(self.loc, "wot"))[0]
        self.copy_wotfile(sigfile)
        self.args.f = c_wot
        n = self.v_main()
        self.assertEquals(len(n), 1)
        # TODO: test for set equality here

    def test_lists_several_wot_members(self):
        for s in listdir(join(self.loc, "wot"))[:2]:
            self.copy_wotfile(s)
        self.args.f = c_wot
        n = self.v_main()
        self.assertEquals(len(n), 2)

    def test_orchestra_antecedents(self):
        self.orchestrate()
        n = self.v_main(
            cmd=c_ante,
            query=join(self.patchdir,
                       "mod6_fix_dumpblock_params.vpatch"))
        self.assertEquals(
            n.keys(),
            [join(self.scratchdir, "patches", "asciilifeform_and_now_we_have_block_dumper_corrected.vpatch")])

    def test_orchestra_descendants(self):
        self.orchestrate()
        known_descs = [
            "asciilifeform_lets_lose_testnet.vpatch",
            "bitcoin-asciilifeform.4-goodbye-win32.vpatch",
            "asciilifeform_dns_thermonyukyoolar_kleansing.vpatch",
            "rm_rf_upnp.vpatch",
            "asciilifeform_ver_now_5_4_and_irc_is_gone_and_now_must_give_ip.vpatch",
            "bitcoin-asciilifeform.1.vpatch"]
        n = self.v_main(cmd=c_desc,
                        query=join(self.patchdir, "genesis.vpatch"))
        self.assertEquals(
            sorted(n.keys()),
            sorted([join(self.scratchdir,
                  "patches",
                  p) for p in known_descs]))

    def test_flow(self):
        self.athenize()
        self.maxDiff = None
        self.assertEquals(
            self.v_main(cmd=c_flow),
            {join(self.scratchdir, "patches", "genesis.vpatch"):
             ["asciilifeform"]})

    def test_flow_with_two_patches(self):
        self.athenize()
        self.maxDiff = None
        self.copy_wotfile("mircea_popescu.asc")
        self.copy_sigfile("genesis.vpatch.mircea_popescu.sig")
        self.assertEquals(
            self.v_main(cmd=c_flow),
            {join(self.scratchdir, "patches", "genesis.vpatch"):
             ["asciilifeform", "mircea_popescu"]})

    def test_origin_with_a_patch(self):
        self.athenize()
        self.assertEquals(
            self.v_main(cmd=c_origin,
                        query="9353b178cfd30e8aead331647bbe3e2cd73039ed2dbb04ef0b2194cec44735bb21624ec938a506f5a7da8a9589799b55cbb3416cc01e6cc16269fed4e9cf19f3"),
            True)

    def tearDown(self):
        rmtree(self.scratchdir)
